#!/usr/bin/env python
# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SETools.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import setools
import argparse
import sys
import logging

parser = argparse.ArgumentParser(
    description="SELinux policy rule search tool.",
    epilog="TE/MLS rule searches cannot be mixed with RBAC rule searches.")
parser.add_argument("--version", action="version", version=setools.__version__)
parser.add_argument("policy", help="Path to the SELinux policy to search.", nargs="?")
parser.add_argument("-v", "--verbose", action="store_true",
                    help="Print extra informational messages")
parser.add_argument("--debug", action="store_true", dest="debug", help="Enable debugging.")

rtypes = parser.add_argument_group("TE Rule Types")
rtypes.add_argument("-A", action="store_true", help="Search allow and allowxperm rules.")
rtypes.add_argument("--allow", action="append_const",
                    const="allow", dest="tertypes",
                    help="Search allow rules.")
rtypes.add_argument("--allowxperm", action="append_const",
                    const="allowxperm", dest="tertypes",
                    help="Search allowxperm rules.")
rtypes.add_argument("--auditallow", action="append_const",
                    const="auditallow", dest="tertypes",
                    help="Search auditallow rules.")
rtypes.add_argument("--auditallowxperm", action="append_const",
                    const="auditallowxperm", dest="tertypes",
                    help="Search auditallowxperm rules.")
rtypes.add_argument("--dontaudit", action="append_const",
                    const="dontaudit", dest="tertypes",
                    help="Search dontaudit rules.")
rtypes.add_argument("--dontauditxperm", action="append_const",
                    const="dontauditxperm", dest="tertypes",
                    help="Search dontauditxperm rules.")
rtypes.add_argument("--neverallow", action="append_const",
                    const="neverallow", dest="tertypes",
                    help="Search neverallow rules.")
rtypes.add_argument("--neverallowxperm", action="append_const",
                    const="neverallowxperm", dest="tertypes",
                    help="Search neverallowxperm rules.")
rtypes.add_argument("-T", "--type_trans", action="append_const",
                    const="type_transition", dest="tertypes",
                    help="Search type_transition rules.")
rtypes.add_argument("--type_change", action="append_const",
                    const="type_change", dest="tertypes",
                    help="Search type_change rules.")
rtypes.add_argument("--type_member", action="append_const",
                    const="type_member", dest="tertypes",
                    help="Search type_member rules.")
rbacrtypes = parser.add_argument_group("RBAC Rule Types")
rbacrtypes.add_argument("--role_allow", action="append_const",
                        const="allow", dest="rbacrtypes",
                        help="Search role allow rules.")
rbacrtypes.add_argument("--role_trans", action="append_const",
                        const="role_transition", dest="rbacrtypes",
                        help="Search role_transition rules.")

mlsrtypes = parser.add_argument_group("MLS Rule Types")
mlsrtypes.add_argument("--range_trans", action="append_const",
                       const="range_transition", dest="mlsrtypes",
                       help="Search range_transition rules.")

expr = parser.add_argument_group("Expressions")
expr.add_argument("-s", "--source",
                  help="Source type/role of the TE/RBAC rule.")
expr.add_argument("-t", "--target",
                  help="Target type/role of the TE/RBAC rule.")
expr.add_argument("-c", "--class", dest="tclass",
                  help="Comma separated list of object classes")
expr.add_argument("-p", "--perms", metavar="PERMS",
                  help="Comma separated list of permissions.")
expr.add_argument("-x", "--xperms", metavar="XPERMS",
                  help="Comma separated list of extended permissions.")
expr.add_argument("-D", "--default",
                  help="Default of the rule. (type/role/range transition rules)")
expr.add_argument("-b", "--bool", dest="boolean", metavar="BOOL",
                  help="Comma separated list of Booleans in the conditional expression.")

opts = parser.add_argument_group("Search options")
opts.add_argument("-eb", action="store_true", dest="boolean_equal",
                  help="Match Boolean list exactly instead of matching any listed Boolean.")
opts.add_argument("-ep", action="store_true", dest="perms_equal",
                  help="Match permission set exactly instead of matching any listed permission.")
opts.add_argument("-ex", action="store_true", dest="xperms_equal",
                  help="Match extended permission set exactly instead of matching any listed "
                  "permission.")
opts.add_argument("-ds", action="store_false", dest="source_indirect",
                  help="Match source attributes directly instead of matching member types/roles.")
opts.add_argument("-dt", action="store_false", dest="target_indirect",
                  help="Match target attributes directly instead of matching member types/roles.")
opts.add_argument("-rs", action="store_true", dest="source_regex",
                  help="Use regular expression matching for the source type/role.")
opts.add_argument("-rt", action="store_true", dest="target_regex",
                  help="Use regular expression matching for the target type/role.")
opts.add_argument("-rc", action="store_true", dest="tclass_regex",
                  help="Use regular expression matching for the object class.")
opts.add_argument("-rd", action="store_true", dest="default_regex",
                  help="Use regular expression matching for the default type/role.")
opts.add_argument("-rb", action="store_true", dest="boolean_regex",
                  help="Use regular expression matching for Booleans.")

args = parser.parse_args()

if args.A:
    try:
        args.tertypes.extend(["allow", "allowxperm"])
    except AttributeError:
        args.tertypes = ["allow", "allowxperm"]

if not args.tertypes and not args.mlsrtypes and not args.rbacrtypes:
    parser.error("At least one rule type must be specified.")

if args.debug:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s|%(levelname)s|%(name)s|%(message)s')
elif args.verbose:
    logging.basicConfig(level=logging.INFO, format='%(message)s')
else:
    logging.basicConfig(level=logging.WARNING, format='%(message)s')

try:
    p = setools.SELinuxPolicy(args.policy)

    if args.tertypes:
        q = setools.TERuleQuery(p,
                                ruletype=args.tertypes,
                                source=args.source,
                                source_indirect=args.source_indirect,
                                source_regex=args.source_regex,
                                target=args.target,
                                target_indirect=args.target_indirect,
                                target_regex=args.target_regex,
                                tclass_regex=args.tclass_regex,
                                perms_equal=args.perms_equal,
                                xperms_equal=args.xperms_equal,
                                default=args.default,
                                default_regex=args.default_regex,
                                boolean_regex=args.boolean_regex,
                                boolean_equal=args.boolean_equal)

        # these are broken out from the above statement to prevent making a list
        # with an empty string in it (split on empty string)
        if args.tclass:
            if args.tclass_regex:
                q.tclass = args.tclass
            else:
                q.tclass = args.tclass.split(",")

        if args.perms:
            q.perms = args.perms.split(",")

        if args.xperms:
            xperms = []
            for item in args.xperms.split(","):
                rng = item.split("-")
                if len(rng) == 2:
                    xperms.append((int(rng[0], base=16), int(rng[1], base=16)))
                elif len(rng) == 1:
                    xperms.append((int(rng[0], base=16), int(rng[0], base=16)))
                else:
                    parser.error("Enter an extended permission or extended permission range, e.g. "
                                 "0x5411 or 0x8800-0x88ff.")

            q.xperms = xperms

        if args.boolean:
            if args.boolean_regex:
                q.boolean = args.boolean
            else:
                q.boolean = args.boolean.split(",")

        for r in sorted(q.results()):
            print(r)

    if args.rbacrtypes:
        q = setools.RBACRuleQuery(p,
                                  ruletype=args.rbacrtypes,
                                  source=args.source,
                                  source_indirect=args.source_indirect,
                                  source_regex=args.source_regex,
                                  target=args.target,
                                  target_indirect=args.target_indirect,
                                  target_regex=args.target_regex,
                                  default=args.default,
                                  default_regex=args.default_regex,
                                  tclass_regex=args.tclass_regex)

        # these are broken out from the above statement to prevent making a list
        # with an empty string in it (split on empty string)
        if args.tclass:
            if args.tclass_regex:
                q.tclass = args.tclass
            else:
                q.tclass = args.tclass.split(",")

        for r in sorted(q.results()):
            print(r)

    if args.mlsrtypes:
        q = setools.MLSRuleQuery(p,
                                 ruletype=args.mlsrtypes,
                                 source=args.source,
                                 source_indirect=args.source_indirect,
                                 source_regex=args.source_regex,
                                 target=args.target,
                                 target_indirect=args.target_indirect,
                                 target_regex=args.target_regex,
                                 tclass_regex=args.tclass_regex,
                                 default=args.default)

        # these are broken out from the above statement to prevent making a list
        # with an empty string in it (split on empty string)
        if args.tclass:
            if args.tclass_regex:
                q.tclass = args.tclass
            else:
                q.tclass = args.tclass.split(",")

        for r in sorted(q.results()):
            print(r)

except Exception as err:
    if args.debug:
        logging.exception(str(err))
    else:
        print(err)

    sys.exit(1)



#!/usr/bin/env python
# Copyright 2015-2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SETools.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import setools
import argparse
import sys
import logging
from itertools import chain

parser = argparse.ArgumentParser(
    description="SELinux policy semantic difference tool.",
    epilog="If no differences are selected, all differences will be printed.")
parser.add_argument("POLICY1", help="Path to the first SELinux policy to diff.", nargs=1)
parser.add_argument("POLICY2", help="Path to the second SELinux policy to diff.", nargs=1)
parser.add_argument("--version", action="version", version=setools.__version__)
parser.add_argument("--stats", action="store_true", help="Display only statistics.")
parser.add_argument("-v", "--verbose", action="store_true",
                    help="Print extra informational messages")
parser.add_argument("--debug", action="store_true", dest="debug", help="Enable debugging.")

comp = parser.add_argument_group("component differences")
comp.add_argument("--common", action="store_true", help="Print common differences")
comp.add_argument("-c", "--class", action="store_true", help="Print class differences",
                  dest="class_")
comp.add_argument("-t", "--type", action="store_true", help="Print type differences",
                  dest="type_")
comp.add_argument("-a", "--attribute", action="store_true", help="Print type attribute differences")
comp.add_argument("-r", "--role", action="store_true", help="Print role differences")
comp.add_argument("-u", "--user", action="store_true", help="Print user differences")
comp.add_argument("-b", "--bool", action="store_true", help="Print Boolean differences",
                  dest="bool_")
comp.add_argument("--sensitivity", action="store_true", help="Print MLS sensitivity differences")
comp.add_argument("--category", action="store_true", help="Print MLS category differences")
comp.add_argument("--level", action="store_true", help="Print MLS level definition differences")

terule = parser.add_argument_group("type enforcement rule differences")
terule.add_argument("-A", action="store_true", help="Print allow and allowxperm rule differences")
terule.add_argument("--allow", action="store_true", help="Print allow rule differences")
terule.add_argument("--neverallow", action="store_true", help="Print neverallow rule differences")
terule.add_argument("--auditallow", action="store_true", help="Print auditallow rule differences")
terule.add_argument("--dontaudit", action="store_true", help="Print dontaudit rule differences")
terule.add_argument("--allowxperm", action="store_true", help="Print allowxperm rule differences")
terule.add_argument("--neverallowxperm", action="store_true",
                    help="Print neverallowxperm rule differences")
terule.add_argument("--auditallowxperm", action="store_true",
                    help="Print auditallowxperm rule differences")
terule.add_argument("--dontauditxperm", action="store_true",
                    help="Print dontauditxperm rule differences")
terule.add_argument("-T", "--type_trans", action="store_true",
                    help="Print type_transition rule differences")
terule.add_argument("--type_change", action="store_true", help="Print type_change rule differences")
terule.add_argument("--type_member", action="store_true",
                    help="Print type_member rule differences")

rbacrule = parser.add_argument_group("RBAC rule differences")
rbacrule.add_argument("--role_allow", action="store_true", help="Print role allow rule differences")
rbacrule.add_argument("--role_trans", action="store_true",
                      help="Print role_transition rule differences")

mlsrule = parser.add_argument_group("MLS rule differences")
mlsrule.add_argument("--range_trans", action="store_true",
                     help="Print range_transition rule differences")

constrain = parser.add_argument_group("Constraint differences")
constrain.add_argument("--constrain", action="store_true", help="Print constrain differences")
constrain.add_argument("--mlsconstrain", action="store_true", help="Print mlsconstrain differences")
constrain.add_argument("--validatetrans", action="store_true",
                       help="Print validatetrans differences")
constrain.add_argument("--mlsvalidatetrans", action="store_true",
                       help="Print mlsvalidatetrans differences")

labeling = parser.add_argument_group("labeling statement differences")
labeling.add_argument("--initialsid", action="store_true", help="Print initial SID differences")
labeling.add_argument("--fs_use", action="store_true", help="Print fs_use_* differences")
labeling.add_argument("--genfscon", action="store_true", help="Print genfscon differences")
labeling.add_argument("--netifcon", action="store_true", help="Print netifcon differences")
labeling.add_argument("--nodecon", action="store_true", help="Print nodecon differences")
labeling.add_argument("--portcon", action="store_true", help="Print portcon differences")

other = parser.add_argument_group("other differences")
other.add_argument("--default", action="store_true", help="Print default_* differences")
other.add_argument("--property", action="store_true",
                   help="Print policy property differences (handle_unknown, version, MLS)")
other.add_argument("--polcap", action="store_true", help="Print policy capability differences")
other.add_argument("--typebounds", action="store_true", help="Print typebounds differences")

args = parser.parse_args()

if args.A:
    args.allow = True
    args.allowxperm = True

all_differences = not any((args.class_, args.common, args.type_, args.attribute, args.role,
                           args.user, args.bool_, args.sensitivity, args.category, args.level,
                           args.allow, args.neverallow, args.auditallow, args.dontaudit,
                           args.type_trans, args.type_change, args.type_member, args.role_allow,
                           args.role_trans, args.range_trans, args.initialsid, args.genfscon,
                           args.netifcon, args.nodecon, args.portcon, args.fs_use, args.polcap,
                           args.property, args.default, args.constrain, args.mlsconstrain,
                           args.validatetrans, args.mlsvalidatetrans, args.typebounds,
                           args.allowxperm, args.neverallowxperm, args.auditallowxperm,
                           args.dontauditxperm))

if args.debug:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s|%(levelname)s|%(name)s|%(message)s')
elif args.verbose:
    logging.basicConfig(level=logging.INFO, format='%(message)s')
else:
    logging.basicConfig(level=logging.WARNING, format='%(message)s')

try:
    p1 = setools.SELinuxPolicy(args.POLICY1[0])
    p2 = setools.SELinuxPolicy(args.POLICY2[0])
    diff = setools.PolicyDifference(p1, p2)

    if all_differences or args.property:
        print("Policy Properties ({0} Modified)".format(len(diff.modified_properties)))

        if diff.modified_properties and not args.stats:
            for name, added, removed in sorted(diff.modified_properties, key=lambda x: x.property):
                print("      * {0} +{1} -{2}".format(name, added, removed))

        print()

    if all_differences or args.common:
        if diff.added_commons or diff.removed_commons or diff.modified_commons or args.common:
            print("Commons ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_commons), len(diff.removed_commons), len(diff.modified_commons)))
            if diff.added_commons and not args.stats:
                print("   Added Commons: {0}".format(len(diff.added_commons)))
                for c in sorted(diff.added_commons):
                    print("      + {0}".format(c))
            if diff.removed_commons and not args.stats:
                print("   Removed Commons: {0}".format(len(diff.removed_commons)))
                for c in sorted(diff.removed_commons):
                    print("      - {0}".format(c))
            if diff.modified_commons and not args.stats:
                print("   Modified Commons: {0}".format(len(diff.modified_commons)))
                for name, mod in sorted(diff.modified_commons.items()):
                    change = []
                    if mod.added_perms:
                        change.append("{0} Added permissions".format(len(mod.added_perms)))
                    if mod.removed_perms:
                        change.append("{0} Removed permissions".format(len(mod.removed_perms)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    for p in sorted(mod.added_perms):
                        print("          + {0}".format(p))
                    for p in sorted(mod.removed_perms):
                        print("          - {0}".format(p))
            print()

    if all_differences or args.class_:
        if diff.added_classes or diff.removed_classes or diff.modified_classes or args.class_:
            print("Classes ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_classes), len(diff.removed_classes), len(diff.modified_classes)))
            if diff.added_classes and not args.stats:
                print("   Added Classes: {0}".format(len(diff.added_classes)))
                for c in sorted(diff.added_classes):
                    print("      + {0}".format(c))
            if diff.removed_classes and not args.stats:
                print("   Removed Classes: {0}".format(len(diff.removed_classes)))
                for c in sorted(diff.removed_classes):
                    print("      - {0}".format(c))
            if diff.modified_classes and not args.stats:
                print("   Modified Classes: {0}".format(len(diff.modified_classes)))
                for name, mod in sorted(diff.modified_classes.items()):
                    change = []
                    if mod.added_perms:
                        change.append("{0} Added permissions".format(len(mod.added_perms)))
                    if mod.removed_perms:
                        change.append("{0} Removed permissions".format(len(mod.removed_perms)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    for p in sorted(mod.added_perms):
                        print("          + {0}".format(p))
                    for p in sorted(mod.removed_perms):
                        print("          - {0}".format(p))
            print()

    if all_differences or args.bool_:
        if diff.added_booleans or diff.removed_booleans or \
                diff.modified_booleans or args.bool_:
            print("Booleans ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_booleans), len(diff.removed_booleans),
                len(diff.modified_booleans)))
            if diff.added_booleans and not args.stats:
                print("   Added Booleans: {0}".format(len(diff.added_booleans)))
                for a in sorted(diff.added_booleans):
                    print("      + {0}".format(a))
            if diff.removed_booleans and not args.stats:
                print("   Removed Booleans: {0}".format(len(diff.removed_booleans)))
                for a in sorted(diff.removed_booleans):
                    print("      - {0}".format(a))
            if diff.modified_booleans and not args.stats:
                print("   Modified Booleans: {0}".format(len(diff.modified_booleans)))
                for name, mod in sorted(diff.modified_booleans.items()):
                    print("      * {0} (Modified default state)".format(name))
                    print("          + {0}".format(mod.added_state))
                    print("          - {0}".format(mod.removed_state))

            print()

    if all_differences or args.role:
        if diff.added_roles or diff.removed_roles or diff.modified_roles or args.role:
            print("Roles ({0} Added, {1} Removed, {2} Modified)".format(len(diff.added_roles),
                                                                        len(diff.removed_roles),
                                                                        len(diff.modified_roles)))
            if diff.added_roles and not args.stats:
                print("   Added Roles: {0}".format(len(diff.added_roles)))
                for r in sorted(diff.added_roles):
                    print("      + {0}".format(r))
            if diff.removed_roles and not args.stats:
                print("   Removed Roles: {0}".format(len(diff.removed_roles)))
                for r in sorted(diff.removed_roles):
                    print("      - {0}".format(r))
            if diff.modified_roles and not args.stats:
                print("   Modified Roles: {0}".format(len(diff.modified_roles)))
                for name, mod in sorted(diff.modified_roles.items()):
                    change = []
                    if mod.added_types:
                        change.append("{0} Added types".format(len(mod.added_types)))
                    if mod.removed_types:
                        change.append("{0} Removed types".format(len(mod.removed_types)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    for t in sorted(mod.added_types):
                        print("          + {0}".format(t))
                    for t in sorted(mod.removed_types):
                        print("          - {0}".format(t))
            print()

    if all_differences or args.type_:
        if diff.added_types or diff.removed_types or diff.modified_types or args.type_:
            print("Types ({0} Added, {1} Removed, {2} Modified)".format(len(diff.added_types),
                                                                        len(diff.removed_types),
                                                                        len(diff.modified_types)))
            if diff.added_types and not args.stats:
                print("   Added Types: {0}".format(len(diff.added_types)))
                for r in sorted(diff.added_types):
                    print("      + {0}".format(r))
            if diff.removed_types and not args.stats:
                print("   Removed Types: {0}".format(len(diff.removed_types)))
                for r in sorted(diff.removed_types):
                    print("      - {0}".format(r))
            if diff.modified_types and not args.stats:
                print("   Modified Types: {0}".format(len(diff.modified_types)))
                for name, mod in sorted(diff.modified_types.items()):
                    change = []
                    if mod.added_attributes:
                        change.append("{0} Added attributes".format(len(mod.added_attributes)))
                    if mod.removed_attributes:
                        change.append("{0} Removed attributes".format(len(mod.removed_attributes)))
                    if mod.added_aliases:
                        change.append("{0} Added aliases".format(len(mod.added_aliases)))
                    if mod.removed_aliases:
                        change.append("{0} Removed aliases".format(len(mod.removed_aliases)))
                    if mod.modified_permissive:
                        if mod.permissive:
                            change.append("Removed permissive")
                        else:
                            change.append("Added permissive")

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    if mod.added_attributes or mod.removed_attributes:
                        print("          Attributes:")
                    for t in sorted(mod.added_attributes):
                        print("          + {0}".format(t))
                    for t in sorted(mod.removed_attributes):
                        print("          - {0}".format(t))

                    if mod.added_aliases or mod.removed_aliases:
                        print("          Aliases:")
                    for t in sorted(mod.added_aliases):
                        print("          + {0}".format(t))
                    for t in sorted(mod.removed_aliases):
                        print("          - {0}".format(t))

            print()

    if all_differences or args.attribute:
        if diff.added_type_attributes or diff.removed_type_attributes or \
                diff.modified_type_attributes or args.attribute:
            print("Type Attributes ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_type_attributes), len(diff.removed_type_attributes),
                len(diff.modified_type_attributes)))
            if diff.added_type_attributes and not args.stats:
                print("   Added Type Attributes: {0}".format(len(diff.added_type_attributes)))
                for a in sorted(diff.added_type_attributes):
                    print("      + {0}".format(a))
            if diff.removed_type_attributes and not args.stats:
                print("   Removed Type Attributes: {0}".format(len(diff.removed_type_attributes)))
                for a in sorted(diff.removed_type_attributes):
                    print("      - {0}".format(a))
            if diff.modified_type_attributes and not args.stats:
                print("   Modified Type Attributes: {0}".format(len(diff.modified_type_attributes)))
                for name, mod in sorted(diff.modified_type_attributes.items()):
                    change = []
                    if mod.added_types:
                        change.append("{0} Added types".format(len(mod.added_types)))
                    if mod.removed_types:
                        change.append("{0} Removed types".format(len(mod.removed_types)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    for t in sorted(mod.added_types):
                        print("          + {0}".format(t))
                    for t in sorted(mod.removed_types):
                        print("          - {0}".format(t))
            print()

    if all_differences or args.user:
        if diff.added_users or diff.removed_users or diff.modified_users or args.user:
            print("Users ({0} Added, {1} Removed, {2} Modified)".format(len(diff.added_users),
                                                                        len(diff.removed_users),
                                                                        len(diff.modified_users)))
            if diff.added_users and not args.stats:
                print("   Added Users: {0}".format(len(diff.added_users)))
                for u in sorted(diff.added_users):
                    print("      + {0}".format(u))
            if diff.removed_users and not args.stats:
                print("   Removed Users: {0}".format(len(diff.removed_users)))
                for u in sorted(diff.removed_users):
                    print("      - {0}".format(u))
            if diff.modified_users and not args.stats:
                print("   Modified Users: {0}".format(len(diff.modified_users)))
                for name, mod in sorted(diff.modified_users.items()):
                    change = []
                    if mod.added_roles:
                        change.append("{0} Added roles".format(len(mod.added_roles)))
                    if mod.removed_roles:
                        change.append("{0} Removed roles".format(len(mod.removed_roles)))
                    if mod.removed_level:
                        change.append("Modified default level")
                    if mod.removed_range:
                        change.append("Modified range")

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    if mod.added_roles or mod.removed_roles:
                        print("          Roles:")
                    for t in sorted(mod.added_roles):
                        print("          + {0}".format(t))
                    for t in sorted(mod.removed_roles):
                        print("          - {0}".format(t))

                    if mod.removed_level:
                        print("          Default level:")
                        print("          + {0}".format(mod.added_level))
                        print("          - {0}".format(mod.removed_level))

                    if mod.removed_range:
                        print("          Range:")
                        print("          + {0}".format(mod.added_range))
                        print("          - {0}".format(mod.removed_range))
            print()

    if all_differences or args.category:
        if diff.added_categories or diff.removed_categories or diff.modified_categories \
                or args.category:
            print("Categories ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_categories), len(diff.removed_categories),
                len(diff.modified_categories)))
            if diff.added_categories and not args.stats:
                print("   Added Categories: {0}".format(len(diff.added_categories)))
                for c in sorted(diff.added_categories):
                    print("      + {0}".format(c))
            if diff.removed_categories and not args.stats:
                print("   Removed Categories: {0}".format(len(diff.removed_categories)))
                for c in sorted(diff.removed_categories):
                    print("      - {0}".format(c))
            if diff.modified_categories and not args.stats:
                print("   Modified Categories: {0}".format(len(diff.modified_categories)))
                for name, mod in sorted(diff.modified_categories.items()):
                    change = []
                    if mod.added_aliases:
                        change.append("{0} Added Aliases".format(len(mod.added_aliases)))
                    if mod.removed_aliases:
                        change.append("{0} Removed Aliases".format(len(mod.removed_aliases)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    print("          Aliases:")
                    for a in sorted(mod.added_aliases):
                        print("          + {0}".format(a))
                    for a in sorted(mod.removed_aliases):
                        print("          - {0}".format(a))

            print()

    if all_differences or args.sensitivity:
        if diff.added_sensitivities or diff.removed_sensitivities or diff.modified_sensitivities \
                or args.sensitivity:
            print("Sensitivities ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_sensitivities), len(diff.removed_sensitivities),
                len(diff.modified_sensitivities)))
            if diff.added_sensitivities and not args.stats:
                print("   Added Sensitivites: {0}".format(len(diff.added_sensitivities)))
                for s in sorted(diff.added_sensitivities):
                    print("      + {0}".format(s))
            if diff.removed_sensitivities and not args.stats:
                print("   Removed Sensitivities: {0}".format(len(diff.removed_sensitivities)))
                for s in sorted(diff.removed_sensitivities):
                    print("      - {0}".format(s))
            if diff.modified_sensitivities and not args.stats:
                print("   Modified Sensitivities: {0}".format(len(diff.modified_sensitivities)))
                for name, mod in sorted(diff.modified_sensitivities.items()):
                    change = []
                    if mod.added_aliases:
                        change.append("{0} Added Aliases".format(len(mod.added_aliases)))
                    if mod.removed_aliases:
                        change.append("{0} Removed Aliases".format(len(mod.removed_aliases)))

                    print("      * {0} ({1})".format(name, ", ".join(change)))
                    print("          Aliases:")
                    for a in sorted(mod.added_aliases):
                        print("          + {0}".format(a))
                    for a in sorted(mod.removed_aliases):
                        print("          - {0}".format(a))

            print()

    if all_differences or args.level:
        if diff.added_levels or diff.removed_levels or \
                diff.modified_levels or args.level:
            print("Levels ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_levels), len(diff.removed_levels),
                len(diff.modified_levels)))
            if diff.added_levels and not args.stats:
                print("   Added Levels: {0}".format(len(diff.added_levels)))
                for l in sorted(diff.added_levels):
                    print("      + {0}".format(l))
            if diff.removed_levels and not args.stats:
                print("   Removed Levels: {0}".format(len(diff.removed_levels)))
                for l in sorted(diff.removed_levels):
                    print("      - {0}".format(l))
            if diff.modified_levels and not args.stats:
                print("   Modified Levels: {0}".format(len(diff.modified_levels)))
                for level, added_categories, removed_categories, _ in sorted(diff.modified_levels,
                                                                             key=lambda x: x.level):
                    change = []
                    if added_categories:
                        change.append("{0} Added Categories".format(len(added_categories)))
                    if removed_categories:
                        change.append("{0} Removed Categories".format(len(removed_categories)))

                    print("      * {0} ({1})".format(level.sensitivity, ", ".join(change)))
                    for c in sorted(added_categories):
                        print("          + {0}".format(c))
                    for c in sorted(removed_categories):
                        print("          - {0}".format(c))
            print()

    if all_differences or args.allow:
        if diff.added_allows or diff.removed_allows or diff.modified_allows or args.allow:
            print("Allow Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_allows), len(diff.removed_allows), len(diff.modified_allows)))

            if diff.added_allows and not args.stats:
                print("   Added Allow Rules: {0}".format(len(diff.added_allows)))
                for r in sorted(diff.added_allows):
                    print("      + {0}".format(r))

            if diff.removed_allows and not args.stats:
                print("   Removed Allow Rules: {0}".format(len(diff.removed_allows)))
                for r in sorted(diff.removed_allows):
                    print("      - {0}".format(r))

            if diff.modified_allows and not args.stats:
                print("   Modified Allow Rules: {0}".format(len(diff.modified_allows)))

                for rule, added_perms, removed_perms, matched_perms in sorted(diff.modified_allows,
                                                                              key=lambda x: x.rule):
                    perms = " ".join(chain((p for p in matched_perms),
                                           ("+" + p for p in added_perms),
                                           ("-" + p for p in removed_perms)))
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} {{ {1} }};".format(
                        rule, perms)

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass
                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.allowxperm:
        if diff.added_allowxperms or diff.removed_allowxperms or diff.modified_allowxperms \
                or args.allowxperm:

            print("Allowxperm Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_allowxperms), len(diff.removed_allowxperms),
                len(diff.modified_allowxperms)))

            if diff.added_allowxperms and not args.stats:
                print("   Added Allowxperm Rules: {0}".format(len(diff.added_allowxperms)))
                for r in sorted(diff.added_allowxperms):
                    print("      + {0}".format(r))

            if diff.removed_allowxperms and not args.stats:
                print("   Removed Allowxperm Rules: {0}".format(len(diff.removed_allowxperms)))
                for r in sorted(diff.removed_allowxperms):
                    print("      - {0}".format(r))

            if diff.modified_allowxperms and not args.stats:
                print("   Modified Allowxperm Rules: {0}".format(len(diff.modified_allowxperms)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_allowxperms, key=lambda x: x.rule):

                    # Process the string representation of the sets
                    # so hex representation and ranges are preserved.
                    # Check if the perm sets have contents, otherwise
                    # split on empty string will be an empty string.
                    # Add brackets to added and removed permissions
                    # in case there is a range of permissions.
                    perms = []
                    if matched_perms:
                        for p in str(matched_perms).split(" "):
                            perms.append(p)
                    if added_perms:
                        for p in str(added_perms).split(" "):
                            if '-' in p:
                                perms.append("+[{0}]".format(p))
                            else:
                                perms.append("+{0}".format(p))
                    if removed_perms:
                        for p in str(removed_perms).split(" "):
                            if '-' in p:
                                perms.append("-[{0}]".format(p))
                            else:
                                perms.append("-{0}".format(p))

                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} {{ {1} }};". \
                        format(rule, perms)

                    print("      * {0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} "
                          "{{ {1} }};".format(rule, " ".join(perms)))

            print()

    if all_differences or args.neverallow:
        if diff.added_neverallows or diff.removed_neverallows or diff.modified_neverallows or \
                args.neverallow:
            print("Neverallow Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_neverallows), len(diff.removed_neverallows),
                len(diff.modified_neverallows)))

            if diff.added_neverallows and not args.stats:
                print("   Added Neverallow Rules: {0}".format(len(diff.added_neverallows)))
                for r in sorted(diff.added_neverallows):
                    print("      + {0}".format(r))

            if diff.removed_neverallows and not args.stats:
                print("   Removed Neverallow Rules: {0}".format(len(diff.removed_neverallows)))
                for r in sorted(diff.removed_neverallows):
                    print("      - {0}".format(r))

            if diff.modified_neverallows and not args.stats:
                print("   Modified Neverallow Rules: {0}".format(len(diff.modified_neverallows)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_neverallows, key=lambda x: x.rule):
                    perms = " ".join(chain((p for p in matched_perms),
                                           ("+" + p for p in added_perms),
                                           ("-" + p for p in removed_perms)))
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} {{ {1} }};".format(
                        rule, perms)

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass
                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.neverallowxperm:
        if diff.added_neverallowxperms or diff.removed_neverallowxperms or \
                diff.modified_neverallowxperms or args.neverallowxperm:

            print("Neverallowxperm Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_neverallowxperms), len(diff.removed_neverallowxperms),
                len(diff.modified_neverallowxperms)))

            if diff.added_neverallowxperms and not args.stats:
                print("   Added Neverallowxperm Rules: {0}".format(
                      len(diff.added_neverallowxperms)))
                for r in sorted(diff.added_neverallowxperms):
                    print("      + {0}".format(r))

            if diff.removed_neverallowxperms and not args.stats:
                print("   Removed Neverallowxperm Rules: {0}".format(
                      len(diff.removed_neverallowxperms)))
                for r in sorted(diff.removed_neverallowxperms):
                    print("      - {0}".format(r))

            if diff.modified_neverallowxperms and not args.stats:
                print("   Modified Neverallowxperm Rules: {0}".format(
                      len(diff.modified_neverallowxperms)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_neverallowxperms, key=lambda x: x.rule):

                    # Process the string representation of the sets
                    # so hex representation and ranges are preserved.
                    # Check if the perm sets have contents, otherwise
                    # split on empty string will be an empty string.
                    # Add brackets to added and removed permissions
                    # in case there is a range of permissions.
                    perms = []
                    if matched_perms:
                        for p in str(matched_perms).split(" "):
                            perms.append(p)
                    if added_perms:
                        for p in str(added_perms).split(" "):
                            if '-' in p:
                                perms.append("+[{0}]".format(p))
                            else:
                                perms.append("+{0}".format(p))
                    if removed_perms:
                        for p in str(removed_perms).split(" "):
                            if '-' in p:
                                perms.append("-[{0}]".format(p))
                            else:
                                perms.append("-{0}".format(p))

                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} {{ {1} }};". \
                        format(rule, perms)

                    print("      * {0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} "
                          "{{ {1} }};".format(rule, " ".join(perms)))

            print()

    if all_differences or args.auditallow:
        if diff.added_auditallows or diff.removed_auditallows or diff.modified_auditallows or \
                args.auditallow:
            print("Auditallow Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_auditallows), len(diff.removed_auditallows),
                len(diff.modified_auditallows)))

            if diff.added_auditallows and not args.stats:
                print("   Added Auditallow Rules: {0}".format(len(diff.added_auditallows)))
                for r in sorted(diff.added_auditallows):
                    print("      + {0}".format(r))

            if diff.removed_auditallows and not args.stats:
                print("   Removed Auditallow Rules: {0}".format(len(diff.removed_auditallows)))
                for r in sorted(diff.removed_auditallows):
                    print("      - {0}".format(r))

            if diff.modified_auditallows and not args.stats:
                print("   Modified Auditallow Rules: {0}".format(len(diff.modified_auditallows)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_auditallows, key=lambda x: x.rule):
                    perms = " ".join(chain((p for p in matched_perms),
                                           ("+" + p for p in added_perms),
                                           ("-" + p for p in removed_perms)))
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} {{ {1} }};".format(
                        rule, perms)

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass
                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.auditallowxperm:
        if diff.added_auditallowxperms or diff.removed_auditallowxperms or \
                diff.modified_auditallowxperms or args.auditallowxperm:

            print("Auditallowxperm Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_auditallowxperms), len(diff.removed_auditallowxperms),
                len(diff.modified_auditallowxperms)))

            if diff.added_auditallowxperms and not args.stats:
                print("   Added Auditallowxperm Rules: {0}".format(
                      len(diff.added_auditallowxperms)))
                for r in sorted(diff.added_auditallowxperms):
                    print("      + {0}".format(r))

            if diff.removed_auditallowxperms and not args.stats:
                print("   Removed Auditallowxperm Rules: {0}".format(
                      len(diff.removed_auditallowxperms)))
                for r in sorted(diff.removed_auditallowxperms):
                    print("      - {0}".format(r))

            if diff.modified_auditallowxperms and not args.stats:
                print("   Modified Auditallowxperm Rules: {0}".format(
                      len(diff.modified_auditallowxperms)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_auditallowxperms, key=lambda x: x.rule):

                    # Process the string representation of the sets
                    # so hex representation and ranges are preserved.
                    # Check if the perm sets have contents, otherwise
                    # split on empty string will be an empty string.
                    # Add brackets to added and removed permissions
                    # in case there is a range of permissions.
                    perms = []
                    if matched_perms:
                        for p in str(matched_perms).split(" "):
                            perms.append(p)
                    if added_perms:
                        for p in str(added_perms).split(" "):
                            if '-' in p:
                                perms.append("+[{0}]".format(p))
                            else:
                                perms.append("+{0}".format(p))
                    if removed_perms:
                        for p in str(removed_perms).split(" "):
                            if '-' in p:
                                perms.append("-[{0}]".format(p))
                            else:
                                perms.append("-{0}".format(p))

                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} {{ {1} }};". \
                        format(rule, perms)

                    print("      * {0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} "
                          "{{ {1} }};".format(rule, " ".join(perms)))

            print()

    if all_differences or args.dontaudit:
        if diff.added_dontaudits or diff.removed_dontaudits or diff.modified_dontaudits or \
                args.dontaudit:
            print("Dontaudit Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_dontaudits), len(diff.removed_dontaudits),
                len(diff.modified_dontaudits)))

            if diff.added_dontaudits and not args.stats:
                print("   Added Dontaudit Rules: {0}".format(len(diff.added_dontaudits)))
                for r in sorted(diff.added_dontaudits):
                    print("      + {0}".format(r))

            if diff.removed_dontaudits and not args.stats:
                print("   Removed Dontaudit Rules: {0}".format(len(diff.removed_dontaudits)))
                for r in sorted(diff.removed_dontaudits):
                    print("      - {0}".format(r))

            if diff.modified_dontaudits and not args.stats:
                print("   Modified Dontaudit Rules: {0}".format(len(diff.modified_dontaudits)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_dontaudits, key=lambda x: x.rule):
                    perms = " ".join(chain((p for p in matched_perms),
                                           ("+" + p for p in added_perms),
                                           ("-" + p for p in removed_perms)))
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} {{ {1} }};".format(
                        rule, perms)

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass
                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.dontauditxperm:
        if diff.added_dontauditxperms or diff.removed_dontauditxperms or \
                diff.modified_dontauditxperms or args.dontauditxperm:

            print("Dontauditxperm Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_dontauditxperms), len(diff.removed_dontauditxperms),
                len(diff.modified_dontauditxperms)))

            if diff.added_dontauditxperms and not args.stats:
                print("   Added Dontauditxperm Rules: {0}".format(
                      len(diff.added_dontauditxperms)))
                for r in sorted(diff.added_dontauditxperms):
                    print("      + {0}".format(r))

            if diff.removed_dontauditxperms and not args.stats:
                print("   Removed Dontauditxperm Rules: {0}".format(
                      len(diff.removed_dontauditxperms)))
                for r in sorted(diff.removed_dontauditxperms):
                    print("      - {0}".format(r))

            if diff.modified_dontauditxperms and not args.stats:
                print("   Modified Dontauditxperm Rules: {0}".format(
                      len(diff.modified_dontauditxperms)))

                for rule, added_perms, removed_perms, matched_perms in sorted(
                        diff.modified_dontauditxperms, key=lambda x: x.rule):

                    # Process the string representation of the sets
                    # so hex representation and ranges are preserved.
                    # Check if the perm sets have contents, otherwise
                    # split on empty string will be an empty string.
                    # Add brackets to added and removed permissions
                    # in case there is a range of permissions.
                    perms = []
                    if matched_perms:
                        for p in str(matched_perms).split(" "):
                            perms.append(p)
                    if added_perms:
                        for p in str(added_perms).split(" "):
                            if '-' in p:
                                perms.append("+[{0}]".format(p))
                            else:
                                perms.append("+{0}".format(p))
                    if removed_perms:
                        for p in str(removed_perms).split(" "):
                            if '-' in p:
                                perms.append("-[{0}]".format(p))
                            else:
                                perms.append("-{0}".format(p))

                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} {{ {1} }};". \
                        format(rule, perms)

                    print("      * {0.ruletype} {0.source} {0.target}:{0.tclass} {0.xperm_type} "
                          "{{ {1} }};".format(rule, " ".join(perms)))

            print()

    if all_differences or args.type_trans:
        if diff.added_type_transitions or diff.removed_type_transitions or \
                diff.modified_type_transitions or args.type_trans:
            print("Type_transition Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_type_transitions), len(diff.removed_type_transitions),
                len(diff.modified_type_transitions)))

            if diff.added_type_transitions and not args.stats:
                print("   Added Type_transition Rules: {0}".format(
                    len(diff.added_type_transitions)))
                for r in sorted(diff.added_type_transitions):
                    print("      + {0}".format(r))

            if diff.removed_type_transitions and not args.stats:
                print("   Removed Type_transition Rules: {0}".format(
                    len(diff.removed_type_transitions)))
                for r in sorted(diff.removed_type_transitions):
                    print("      - {0}".format(r))

            if diff.modified_type_transitions and not args.stats:
                print("   Modified Type_transition Rules: {0}".format(
                    len(diff.modified_type_transitions)))

                for rule, added_default, removed_default in sorted(diff.modified_type_transitions,
                                                                   key=lambda x: x.rule):
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} +{1} -{2}".format(
                        rule, added_default, removed_default)

                    try:
                        rule_string += " {0}".format(rule.filename)
                    except AttributeError:
                        pass

                    rule_string += ";"

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass

                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.type_change:
        if diff.added_type_changes or diff.removed_type_changes or \
                diff.modified_type_changes or args.type_change:
            print("Type_change Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_type_changes), len(diff.removed_type_changes),
                len(diff.modified_type_changes)))

            if diff.added_type_changes and not args.stats:
                print("   Added Type_change Rules: {0}".format(len(diff.added_type_changes)))
                for r in sorted(diff.added_type_changes):
                    print("      + {0}".format(r))

            if diff.removed_type_changes and not args.stats:
                print("   Removed Type_change Rules: {0}".format(len(diff.removed_type_changes)))
                for r in sorted(diff.removed_type_changes):
                    print("      - {0}".format(r))

            if diff.modified_type_changes and not args.stats:
                print("   Modified Type_change Rules: {0}".format(len(diff.modified_type_changes)))

                for rule, added_default, removed_default in sorted(diff.modified_type_changes,
                                                                   key=lambda x: x.rule):
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} +{1} -{2}".format(
                        rule, added_default, removed_default)

                    try:
                        rule_string += " {0}".format(rule.filename)
                    except AttributeError:
                        pass

                    rule_string += ";"

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass

                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.type_member:
        if diff.added_type_members or diff.removed_type_members or \
                diff.modified_type_members or args.type_member:
            print("Type_member Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_type_members), len(diff.removed_type_members),
                len(diff.modified_type_members)))

            if diff.added_type_members and not args.stats:
                print("   Added Type_member Rules: {0}".format(len(diff.added_type_members)))
                for r in sorted(diff.added_type_members):
                    print("      + {0}".format(r))

            if diff.removed_type_members and not args.stats:
                print("   Removed Type_member Rules: {0}".format(len(diff.removed_type_members)))
                for r in sorted(diff.removed_type_members):
                    print("      - {0}".format(r))

            if diff.modified_type_members and not args.stats:
                print("   Modified Type_member Rules: {0}".format(len(diff.modified_type_members)))

                for rule, added_default, removed_default in sorted(diff.modified_type_members,
                                                                   key=lambda x: x.rule):
                    rule_string = "{0.ruletype} {0.source} {0.target}:{0.tclass} +{1} -{2}".format(
                        rule, added_default, removed_default)

                    try:
                        rule_string += " {0}".format(rule.filename)
                    except AttributeError:
                        pass

                    rule_string += ";"

                    try:
                        rule_string += " [ {0} ]".format(rule.conditional)
                    except AttributeError:
                        pass

                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.role_allow:
        if diff.added_role_allows or diff.removed_role_allows or args.role_allow:
            print("Role allow Rules ({0} Added, {1} Removed)".format(
                len(diff.added_role_allows), len(diff.removed_role_allows)))

            if diff.added_role_allows and not args.stats:
                print("   Added Role Allow Rules: {0}".format(
                    len(diff.added_role_allows)))
                for r in sorted(diff.added_role_allows):
                    print("      + {0}".format(r))

            if diff.removed_role_allows and not args.stats:
                print("   Removed Role Allow Rules: {0}".format(
                    len(diff.removed_role_allows)))
                for r in sorted(diff.removed_role_allows):
                    print("      - {0}".format(r))

            print()

    if all_differences or args.role_trans:
        if diff.added_role_transitions or diff.removed_role_transitions or \
                diff.modified_role_transitions or args.role_trans:
            print("Role_transition Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_role_transitions), len(diff.removed_role_transitions),
                len(diff.modified_role_transitions)))

            if diff.added_role_transitions and not args.stats:
                print("   Added Role_transition Rules: {0}".format(
                    len(diff.added_role_transitions)))
                for r in sorted(diff.added_role_transitions):
                    print("      + {0}".format(r))

            if diff.removed_role_transitions and not args.stats:
                print("   Removed Role_transition Rules: {0}".format(
                    len(diff.removed_role_transitions)))
                for r in sorted(diff.removed_role_transitions):
                    print("      - {0}".format(r))

            if diff.modified_role_transitions and not args.stats:
                print("   Modified Role_transition Rules: {0}".format(
                    len(diff.modified_role_transitions)))

                for rule, added_default, removed_default in sorted(diff.modified_role_transitions,
                                                                   key=lambda x: x.rule):
                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} +{1} -{2}".format(
                            rule, added_default, removed_default)

                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.range_trans:
        if diff.added_range_transitions or diff.removed_range_transitions or \
                diff.modified_range_transitions or args.range_trans:
            print("Range_transition Rules ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_range_transitions), len(diff.removed_range_transitions),
                len(diff.modified_range_transitions)))

            if diff.added_range_transitions and not args.stats:
                print("   Added Range_transition Rules: {0}".format(
                    len(diff.added_range_transitions)))
                for r in sorted(diff.added_range_transitions):
                    print("      + {0}".format(r))

            if diff.removed_range_transitions and not args.stats:
                print("   Removed Range_transition Rules: {0}".format(
                    len(diff.removed_range_transitions)))
                for r in sorted(diff.removed_range_transitions):
                    print("      - {0}".format(r))

            if diff.modified_range_transitions and not args.stats:
                print("   Modified Range_transition Rules: {0}".format(
                    len(diff.modified_range_transitions)))

                for rule, added_default, removed_default in sorted(diff.modified_range_transitions,
                                                                   key=lambda x: x.rule):
                    # added brackets around range change for clarity since ranges
                    # can have '-' and spaces.
                    rule_string = \
                        "{0.ruletype} {0.source} {0.target}:{0.tclass} +[{1}] -[{2}]".format(
                            rule, added_default, removed_default)

                    print("      * {0}".format(rule_string))

            print()

    if all_differences or args.constrain:
        if diff.added_constrains or diff.removed_constrains or args.constrain:
            print("Constraints ({0} Added, {1} Removed)".format(
                len(diff.added_constrains), len(diff.removed_constrains)))

            if diff.added_constrains and not args.stats:
                print("   Added Constraints: {0}".format(
                    len(diff.added_constrains)))
                for r in sorted(diff.added_constrains):
                    print("      + {0}".format(r))

            if diff.removed_constrains and not args.stats:
                print("   Removed Constraints: {0}".format(
                    len(diff.removed_constrains)))
                for r in sorted(diff.removed_constrains):
                    print("      - {0}".format(r))

            print()

    if all_differences or args.mlsconstrain:
        if diff.added_mlsconstrains or diff.removed_mlsconstrains or args.mlsconstrain:
            print("MLS Constraints ({0} Added, {1} Removed)".format(
                len(diff.added_mlsconstrains), len(diff.removed_mlsconstrains)))

            if diff.added_mlsconstrains and not args.stats:
                print("   Added MLS Constraints: {0}".format(
                    len(diff.added_mlsconstrains)))
                for r in sorted(diff.added_mlsconstrains):
                    print("      + {0}".format(r))

            if diff.removed_mlsconstrains and not args.stats:
                print("   Removed MLS Constraints: {0}".format(
                    len(diff.removed_mlsconstrains)))
                for r in sorted(diff.removed_mlsconstrains):
                    print("      - {0}".format(r))

            print()

    if all_differences or args.validatetrans:
        if diff.added_validatetrans or diff.removed_validatetrans or args.validatetrans:
            print("Validatetrans ({0} Added, {1} Removed)".format(
                len(diff.added_validatetrans), len(diff.removed_validatetrans)))

            if diff.added_validatetrans and not args.stats:
                print("   Added Validatetrans: {0}".format(
                    len(diff.added_validatetrans)))
                for r in sorted(diff.added_validatetrans):
                    print("      + {0}".format(r))

            if diff.removed_validatetrans and not args.stats:
                print("   Removed Validatetrans: {0}".format(
                    len(diff.removed_validatetrans)))
                for r in sorted(diff.removed_validatetrans):
                    print("      - {0}".format(r))

            print()

    if all_differences or args.mlsvalidatetrans:
        if diff.added_mlsvalidatetrans or diff.removed_mlsvalidatetrans or args.mlsvalidatetrans:
            print("MLS Validatetrans ({0} Added, {1} Removed)".format(
                len(diff.added_mlsvalidatetrans), len(diff.removed_mlsvalidatetrans)))

            if diff.added_mlsvalidatetrans and not args.stats:
                print("   Added MLS Validatetrans: {0}".format(
                    len(diff.added_mlsvalidatetrans)))
                for r in sorted(diff.added_mlsvalidatetrans):
                    print("      + {0}".format(r))

            if diff.removed_mlsvalidatetrans and not args.stats:
                print("   Removed MLS Validatetrans: {0}".format(
                    len(diff.removed_mlsvalidatetrans)))
                for r in sorted(diff.removed_mlsvalidatetrans):
                    print("      - {0}".format(r))

            print()

    if all_differences or args.initialsid:
        if diff.added_initialsids or diff.removed_initialsids or diff.modified_initialsids \
                or args.initialsid:
            print("Initial SIDs ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_initialsids), len(diff.removed_initialsids),
                len(diff.modified_initialsids)))
            if diff.added_initialsids and not args.stats:
                print("   Added Initial SIDs: {0}".format(len(diff.added_initialsids)))
                for s in sorted(diff.added_initialsids):
                    print("      + {0}".format(s.statement()))
            if diff.removed_initialsids and not args.stats:
                print("   Removed Initial SIDs: {0}".format(len(diff.removed_initialsids)))
                for s in sorted(diff.removed_initialsids):
                    print("      - {0}".format(s.statement()))
            if diff.modified_initialsids and not args.stats:
                print("   Modified Initial SIDs: {0}".format(len(diff.modified_initialsids)))
                for name, mod in sorted(diff.modified_initialsids.items()):
                    print("      * {0} +[{1.added_context}] -[{1.removed_context}];".format(
                          name, mod))

            print()

    if all_differences or args.fs_use:
        if diff.added_fs_uses or diff.removed_fs_uses or diff.modified_fs_uses \
                or args.fs_use:
            print("Fs_use ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_fs_uses), len(diff.removed_fs_uses),
                len(diff.modified_fs_uses)))
            if diff.added_fs_uses and not args.stats:
                print("   Added Fs_use: {0}".format(len(diff.added_fs_uses)))
                for s in sorted(diff.added_fs_uses):
                    print("      + {0}".format(s))
            if diff.removed_fs_uses and not args.stats:
                print("   Removed Fs_use: {0}".format(len(diff.removed_fs_uses)))
                for s in sorted(diff.removed_fs_uses):
                    print("      - {0}".format(s))
            if diff.modified_fs_uses and not args.stats:
                print("   Modified Fs_use: {0}".format(len(diff.modified_fs_uses)))
                for entry in sorted(diff.modified_fs_uses, key=lambda x: x.rule):
                    print("      * {0.ruletype} {0.fs} +[{1}] -[{2}];".format(
                          entry.rule, entry.added_context, entry.removed_context))

            print()

    if all_differences or args.genfscon:
        if diff.added_genfscons or diff.removed_genfscons or diff.modified_genfscons \
                or args.genfscon:
            print("Genfscons ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_genfscons), len(diff.removed_genfscons),
                len(diff.modified_genfscons)))
            if diff.added_genfscons and not args.stats:
                print("   Added Genfscons: {0}".format(len(diff.added_genfscons)))
                for s in sorted(diff.added_genfscons):
                    print("      + {0}".format(s))
            if diff.removed_genfscons and not args.stats:
                print("   Removed Genfscons: {0}".format(len(diff.removed_genfscons)))
                for s in sorted(diff.removed_genfscons):
                    print("      - {0}".format(s))
            if diff.modified_genfscons and not args.stats:
                print("   Modified Genfscons: {0}".format(len(diff.modified_genfscons)))
                for entry in sorted(diff.modified_genfscons, key=lambda x: x.rule):
                    print("      * genfscon {0.fs} {0.path} {0.filetype} +[{1}] -[{2}];".format(
                          entry.rule, entry.added_context, entry.removed_context))

            print()

    if all_differences or args.netifcon:
        if diff.added_netifcons or diff.removed_netifcons or \
                diff.modified_netifcons or args.netifcon:
            print("Netifcons ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_netifcons), len(diff.removed_netifcons),
                len(diff.modified_netifcons)))
            if diff.added_netifcons and not args.stats:
                print("   Added Netifcons: {0}".format(len(diff.added_netifcons)))
                for n in sorted(diff.added_netifcons):
                    print("      + {0}".format(n))
            if diff.removed_netifcons and not args.stats:
                print("   Removed Netifcons: {0}".format(len(diff.removed_netifcons)))
                for n in sorted(diff.removed_netifcons):
                    print("      - {0}".format(n))
            if diff.modified_netifcons and not args.stats:
                print("   Modified Netifcons: {0}".format(len(diff.modified_netifcons)))
                for entry in sorted(diff.modified_netifcons, key=lambda x: x.rule):
                    # This output is different than other statements because
                    # it becomes difficult to read if this was condensed
                    # into a single line, especially if both contexts
                    # are modified.
                    change = []
                    if entry.removed_context:
                        change.append("Modified Context")
                    if entry.removed_packet:
                        change.append("Modified Packet Context")

                    print("      * {0.netif} ({1})".format(entry.rule, ", ".join(change)))

                    if entry.removed_context:
                        print("          Context:")
                        print("             + {0}".format(entry.added_context))
                        print("             - {0}".format(entry.removed_context))
                    if entry.removed_packet:
                        print("          Packet Context:")
                        print("             + {0}".format(entry.added_packet))
                        print("             - {0}".format(entry.removed_packet))

            print()

    if all_differences or args.nodecon:
        if diff.added_nodecons or diff.removed_nodecons or diff.modified_nodecons \
                or args.nodecon:
            print("Nodecons ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_nodecons), len(diff.removed_nodecons),
                len(diff.modified_nodecons)))
            if diff.added_nodecons and not args.stats:
                print("   Added Nodecons: {0}".format(len(diff.added_nodecons)))
                for n in sorted(diff.added_nodecons):
                    print("      + {0}".format(n))
            if diff.removed_nodecons and not args.stats:
                print("   Removed Nodecons: {0}".format(len(diff.removed_nodecons)))
                for n in sorted(diff.removed_nodecons):
                    print("      - {0}".format(n))
            if diff.modified_nodecons and not args.stats:
                print("   Modified Nodecons: {0}".format(len(diff.modified_nodecons)))
                for con, added_context, removed_context in sorted(diff.modified_nodecons,
                                                                  key=lambda x: x.rule):
                    print("      * nodecon {0.address} {0.netmask} +[{1}] -[{2}];".format(
                          con, added_context, removed_context))

            print()

    if all_differences or args.portcon:
        if diff.added_portcons or diff.removed_portcons or diff.modified_portcons \
                or args.portcon:
            print("Portcons ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_portcons), len(diff.removed_portcons),
                len(diff.modified_portcons)))
            if diff.added_portcons and not args.stats:
                print("   Added Portcons: {0}".format(len(diff.added_portcons)))
                for n in sorted(diff.added_portcons):
                    print("      + {0}".format(n))
            if diff.removed_portcons and not args.stats:
                print("   Removed Portcons: {0}".format(len(diff.removed_portcons)))
                for n in sorted(diff.removed_portcons):
                    print("      - {0}".format(n))
            if diff.modified_portcons and not args.stats:
                print("   Modified Portcons: {0}".format(len(diff.modified_portcons)))
                for con, added_context, removed_context in sorted(diff.modified_portcons,
                                                                  key=lambda x: x.rule):
                    low, high = con.ports
                    if low == high:
                        print("      * portcon {0.protocol} {1} +[{2}] -[{3}];".format(
                              con, low, added_context, removed_context))
                    else:
                        print("      * portcon {0.protocol} {1}-{2} +[{3}] -[{4}];".format(
                              con, low, high, added_context, removed_context))

            print()

    if all_differences or args.polcap:
        if diff.added_polcaps or diff.removed_polcaps or args.polcap:
            print("Policy Capabilities ({0} Added, {1} Removed)".format(
                len(diff.added_polcaps), len(diff.removed_polcaps)))
            if diff.added_polcaps and not args.stats:
                print("   Added Policy Capabilities: {0}".format(len(diff.added_polcaps)))
                for n in sorted(diff.added_polcaps):
                    print("      + {0}".format(n))
            if diff.removed_polcaps and not args.stats:
                print("   Removed Policy Capabilities: {0}".format(len(diff.removed_polcaps)))
                for n in sorted(diff.removed_polcaps):
                    print("      - {0}".format(n))

            print()

    if all_differences or args.default:
        if diff.added_defaults or diff.removed_defaults or args.default:
            print("Defaults ({0} Added, {1} Removed, {2} Modified)".format(
                len(diff.added_defaults), len(diff.removed_defaults), len(diff.modified_defaults)))
            if diff.added_defaults and not args.stats:
                print("   Added Defaults: {0}".format(len(diff.added_defaults)))
                for d in sorted(diff.added_defaults):
                    print("      + {0}".format(d))
            if diff.removed_defaults and not args.stats:
                print("   Removed Defaults: {0}".format(len(diff.removed_defaults)))
                for d in sorted(diff.removed_defaults):
                    print("      - {0}".format(d))
            if diff.modified_defaults and not args.stats:
                print("   Modified Defaults: {0}".format(len(diff.modified_defaults)))
                for default, added_default, removed_default, added_range, removed_range in sorted(
                        diff.modified_defaults, key=lambda x: x.rule):
                    line = "      * {0.ruletype} {0.tclass} ".format(default)
                    if removed_default:
                        line += "+{0} -{1}".format(added_default, removed_default)
                    else:
                        line += default.default

                    if removed_range:
                        line += " +{0} -{1};".format(added_range, removed_range)
                    else:
                        try:
                            line += " {0};".format(default.default_range)
                        except AttributeError:
                            line += ";"
                    print(line)

            print()

    if all_differences or args.typebounds:
        if diff.added_typebounds or diff.removed_typebounds or args.typebounds:
            print("Typebounds ({0} Added, {1} Removed, {2} Modified)".format(
                  len(diff.added_typebounds), len(diff.removed_typebounds),
                  len(diff.modified_typebounds)))
            if diff.added_typebounds and not args.stats:
                print("   Added Typebounds: {0}".format(len(diff.added_typebounds)))
                for d in sorted(diff.added_typebounds):
                    print("      + {0}".format(d))
            if diff.removed_typebounds and not args.stats:
                print("   Removed Typebounds: {0}".format(len(diff.removed_typebounds)))
                for d in sorted(diff.removed_typebounds):
                    print("      - {0}".format(d))
            if diff.modified_typebounds and not args.stats:
                print("   Modified Typebounds: {0}".format(len(diff.modified_typebounds)))
                for bound, added_bound, removed_bound in sorted(
                        diff.modified_typebounds, key=lambda x: x.rule):
                    print("      * {0.ruletype} +{1} -{2} {0.child};".format(
                          bound, added_bound, removed_bound))

            print()

except Exception as err:
    if args.debug:
        logging.exception(str(err))
    else:
        print(err)

    sys.exit(1)



#!/usr/bin/env python
# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SETools.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import setools
import argparse
import sys
import logging


def expand_attr(attr):
    """Render type and role attributes."""
    items = "\n\t".join(sorted(str(i) for i in attr.expand()))
    contents = items if items else "<empty attribute>"
    return "{0}\n\t{1}".format(attr.statement(), contents)

parser = argparse.ArgumentParser(
    description="SELinux policy information tool.")
parser.add_argument("--version", action="version", version=setools.__version__)
parser.add_argument("policy", help="Path to the SELinux policy to query.", nargs="?")
parser.add_argument("-x", "--expand", action="store_true",
                    help="Print additional information about the specified components.")
parser.add_argument("--flat",  help="Print without item count nor indentation.",
                    dest="flat", default=False, action="store_true")
parser.add_argument("-v", "--verbose", action="store_true",
                    help="Print extra informational messages")
parser.add_argument("--debug", action="store_true", dest="debug", help="Enable debugging.")

queries = parser.add_argument_group("Component Queries")
queries.add_argument("-a", "--attribute",  help="Print type attributes.", dest="typeattrquery",
                     nargs='?', const=True, metavar="ATTR")
queries.add_argument("-b", "--bool", help="Print Booleans.", dest="boolquery",
                     nargs='?', const=True, metavar="BOOL")
queries.add_argument("-c", "--class", help="Print object classes.", dest="classquery",
                     nargs='?', const=True, metavar="CLASS")
queries.add_argument("-r", "--role", help="Print roles.", dest="rolequery",
                     nargs='?', const=True, metavar="ROLE")
queries.add_argument("-t", "--type", help="Print types.", dest="typequery",
                     nargs='?', const=True, metavar="TYPE")
queries.add_argument("-u", "--user", help="Print users.", dest="userquery",
                     nargs='?', const=True, metavar="USER")
queries.add_argument("--category", help="Print MLS categories.", dest="mlscatsquery",
                     nargs='?', const=True, metavar="CAT")
queries.add_argument("--common", help="Print common permission set.", dest="commonquery",
                     nargs='?', const=True, metavar="COMMON")
queries.add_argument("--constrain", help="Print constraints.", dest="constraintquery",
                     nargs='?', const=True, metavar="CLASS")
queries.add_argument("--default", help="Print default_* rules.", dest="defaultquery",
                     nargs='?', const=True, metavar="CLASS")
queries.add_argument("--fs_use", help="Print fs_use statements.", dest="fsusequery",
                     nargs='?', const=True, metavar="FS_TYPE")
queries.add_argument("--genfscon", help="Print genfscon statements.", dest="genfsconquery",
                     nargs='?', const=True, metavar="FS_TYPE")
queries.add_argument("--initialsid", help="Print initial SIDs (contexts).", dest="initialsidquery",
                     nargs='?', const=True, metavar="NAME")
queries.add_argument("--netifcon", help="Print netifcon statements.", dest="netifconquery",
                     nargs='?', const=True, metavar="DEVICE")
queries.add_argument("--nodecon", help="Print nodecon statements.", dest="nodeconquery",
                     nargs='?', const=True, metavar="ADDR")
queries.add_argument("--permissive", help="Print permissive types.", dest="permissivequery",
                     nargs='?', const=True, metavar="TYPE")
queries.add_argument("--polcap", help="Print policy capabilities.", dest="polcapquery",
                     nargs='?', const=True, metavar="NAME")
queries.add_argument("--portcon", help="Print portcon statements.", dest="portconquery",
                     nargs='?', const=True, metavar="PORTNUM[-PORTNUM]")
queries.add_argument("--sensitivity", help="Print MLS sensitivities.", dest="mlssensquery",
                     nargs='?', const=True, metavar="SENS")
queries.add_argument("--typebounds", help="Print typebounds statements.", dest="typeboundsquery",
                     nargs='?', const=True, metavar="BOUND_TYPE")
queries.add_argument("--validatetrans", help="Print validatetrans.", dest="validatetransquery",
                     nargs='?', const=True, metavar="CLASS")
queries.add_argument("--all", help="Print all of the above.  On a Xen policy, the Xen components "
                     "will also be printed", dest="all", default=False, action="store_true")

xen = parser.add_argument_group("Xen Component Queries")
xen.add_argument("--ioportcon", help="Print all ioportcon statements.", dest="ioportconquery",
                 default=False, action="store_true")
xen.add_argument("--iomemcon", help="Print all iomemcon statements.", dest="iomemconquery",
                 default=False, action="store_true")
xen.add_argument("--pcidevicecon", help="Print all pcidevicecon statements.",
                 dest="pcideviceconquery", default=False, action="store_true")
xen.add_argument("--pirqcon", help="Print all pirqcon statements.", dest="pirqconquery",
                 default=False, action="store_true")
xen.add_argument("--devicetreecon", help="Print all devicetreecon statements.",
                 dest="devicetreeconquery", default=False, action="store_true")


args = parser.parse_args()

if args.debug:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s|%(levelname)s|%(name)s|%(message)s')
elif args.verbose:
    logging.basicConfig(level=logging.INFO, format='%(message)s')
else:
    logging.basicConfig(level=logging.WARNING, format='%(message)s')

try:
    p = setools.SELinuxPolicy(args.policy)
    components = []

    if args.boolquery or args.all:
        q = setools.BoolQuery(p)
        if isinstance(args.boolquery, str):
            q.name = args.boolquery

        components.append(("Booleans", q, lambda x: x.statement()))

    if args.mlscatsquery or args.all:
        q = setools.CategoryQuery(p)
        if isinstance(args.mlscatsquery, str):
            q.name = args.mlscatsquery

        components.append(("Categories", q, lambda x: x.statement()))

    if args.classquery or args.all:
        q = setools.ObjClassQuery(p)
        if isinstance(args.classquery, str):
            q.name = args.classquery

        components.append(("Classes", q, lambda x: x.statement()))

    if args.commonquery or args.all:
        q = setools.CommonQuery(p)
        if isinstance(args.commonquery, str):
            q.name = args.commonquery

        components.append(("Commons", q, lambda x: x.statement()))

    if args.constraintquery or args.all:
        q = setools.ConstraintQuery(p, ruletype=["constrain", "mlsconstrain"])
        if isinstance(args.constraintquery, str):
            q.tclass = [args.constraintquery]

        components.append(("Constraints", q, lambda x: x.statement()))

    if args.defaultquery or args.all:
        q = setools.DefaultQuery(p)
        if isinstance(args.defaultquery, str):
            q.tclass = [args.defaultquery]

        components.append(("Default rules", q, lambda x: x.statement()))

    if args.fsusequery or args.all:
        q = setools.FSUseQuery(p)
        if isinstance(args.fsusequery, str):
            q.fs = args.fsusequery

        components.append(("Fs_use", q, lambda x: x.statement()))

    if args.genfsconquery or args.all:
        q = setools.GenfsconQuery(p)
        if isinstance(args.genfsconquery, str):
            q.fs = args.genfsconquery

        components.append(("Genfscon", q, lambda x: x.statement()))

    if args.initialsidquery or args.all:
        q = setools.InitialSIDQuery(p)
        if isinstance(args.initialsidquery, str):
            q.name = args.initialsidquery

        components.append(("Initial SIDs", q, lambda x: x.statement()))

    if args.netifconquery or args.all:
        q = setools.NetifconQuery(p)
        if isinstance(args.netifconquery, str):
            q.name = args.netifconquery

        components.append(("Netifcon", q, lambda x: x.statement()))

    if args.nodeconquery or args.all:
        q = setools.NodeconQuery(p)
        if isinstance(args.nodeconquery, str):
            q.network = args.nodeconquery

        components.append(("Nodecon", q, lambda x: x.statement()))

    if args.permissivequery or args.all:
        q = setools.TypeQuery(p, permissive=True, match_permissive=True)
        if isinstance(args.permissivequery, str):
            q.name = args.permissivequery

        components.append(("Permissive Types", q, lambda x: x.statement()))

    if args.polcapquery or args.all:
        q = setools.PolCapQuery(p)
        if isinstance(args.polcapquery, str):
            q.name = args.polcapquery

        components.append(("Polcap", q, lambda x: x.statement()))

    if args.portconquery or args.all:
        q = setools.PortconQuery(p)
        if isinstance(args.portconquery, str):
            try:
                ports = [int(i) for i in args.portconquery.split("-")]
            except ValueError:
                parser.error("Enter a port number or range, e.g. 22 or 6000-6020")

            if len(ports) == 2:
                q.ports = ports
            elif len(ports) == 1:
                q.ports = (ports[0], ports[0])
            else:
                parser.error("Enter a port number or range, e.g. 22 or 6000-6020")

        components.append(("Portcon", q, lambda x: x.statement()))

    if args.rolequery or args.all:
        q = setools.RoleQuery(p)
        if isinstance(args.rolequery, str):
            q.name = args.rolequery

        components.append(("Roles", q, lambda x: x.statement()))

    if args.mlssensquery or args.all:
        q = setools.SensitivityQuery(p)
        if isinstance(args.mlssensquery, str):
            q.name = args.mlssensquery

        components.append(("Sensitivities", q, lambda x: x.statement()))

    if args.typeboundsquery or args.all:
        q = setools.BoundsQuery(p, ruletype=["typebounds"])
        if isinstance(args.typeboundsquery, str):
            q.child = args.typeboundsquery

        components.append(("Typebounds", q, lambda x: x.statement()))

    if args.typequery or args.all:
        q = setools.TypeQuery(p)
        if isinstance(args.typequery, str):
            q.name = args.typequery

        components.append(("Types", q, lambda x: x.statement()))

    if args.typeattrquery or args.all:
        q = setools.TypeAttributeQuery(p)
        if isinstance(args.typeattrquery, str):
            q.name = args.typeattrquery

        components.append(("Type Attributes", q, expand_attr))

    if args.userquery or args.all:
        q = setools.UserQuery(p)
        if isinstance(args.userquery, str):
            q.name = args.userquery

        components.append(("Users", q, lambda x: x.statement()))

    if args.validatetransquery or args.all:
        q = setools.ConstraintQuery(p, ruletype=["validatetrans", "mlsvalidatetrans"])
        if isinstance(args.validatetransquery, str):
            q.tclass = [args.validatetransquery]

        components.append(("Validatetrans", q, lambda x: x.statement()))

    if p.target_platform == "xen":
        if args.ioportconquery or args.all:
            q = setools.IoportconQuery(p)
            components.append(("Ioportcon", q, lambda x: x.statement()))

        if args.iomemconquery or args.all:
            q = setools.IomemconQuery(p)
            components.append(("Iomemcon", q, lambda x: x.statement()))

        if args.pcideviceconquery or args.all:
            q = setools.PcideviceconQuery(p)
            components.append(("Pcidevicecon", q, lambda x: x.statement()))

        if args.pirqconquery or args.all:
            q = setools.PirqconQuery(p)
            components.append(("Pirqcon", q, lambda x: x.statement()))

        if args.devicetreeconquery or args.all:
            q = setools.DevicetreeconQuery(p)
            components.append(("Devicetreecon", q, lambda x: x.statement()))

    if (not components or args.all) and not args.flat:
        mls = "enabled" if p.mls else "disabled"

        print("Statistics for policy file: {0}".format(p))
        print("Policy Version:             {0} (MLS {1})".format(p.version, mls))
        print("Target Policy:              {0}".format(p.target_platform))
        print("Handle unknown classes:     {0}".format(p.handle_unknown))
        print("  Classes:         {0:7}    Permissions:     {1:7}".format(
            p.class_count, p.permission_count))
        print("  Sensitivities:   {0:7}    Categories:      {1:7}".format(
            p.level_count, p.category_count))
        print("  Types:           {0:7}    Attributes:      {1:7}".format(
            p.type_count, p.type_attribute_count))
        print("  Users:           {0:7}    Roles:           {1:7}".format(
            p.user_count, p.role_count))
        print("  Booleans:        {0:7}    Cond. Expr.:     {1:7}".format(
            p.boolean_count, p.conditional_count))
        print("  Allow:           {0:7}    Neverallow:      {1:7}".format(
            p.allow_count, p.neverallow_count))
        print("  Auditallow:      {0:7}    Dontaudit:       {1:7}".format(
            p.auditallow_count, p.dontaudit_count))
        print("  Type_trans:      {0:7}    Type_change:     {1:7}".format(
            p.type_transition_count, p.type_change_count))
        print("  Type_member:     {0:7}    Range_trans:     {1:7}".format(
            p.type_member_count, p.range_transition_count))
        print("  Role allow:      {0:7}    Role_trans:      {1:7}".format(
            p.role_allow_count, p.role_transition_count))
        print("  Constraints:     {0:7}    Validatetrans:   {1:7}".format(
            p.constraint_count, p.validatetrans_count))
        print("  MLS Constrain:   {0:7}    MLS Val. Tran:   {1:7}".format(
            p.mlsconstraint_count, p.mlsvalidatetrans_count))
        print("  Permissives:     {0:7}    Polcap:          {1:7}".format(
            p.permissives_count, p.polcap_count))
        print("  Defaults:        {0:7}    Typebounds:      {1:7}".format(
            p.default_count, p.typebounds_count))

        if p.target_platform == "selinux":
            print("  Allowxperm:      {0:7}    Neverallowxperm: {1:7}".format(
                p.allowxperm_count, p.neverallowxperm_count))
            print("  Auditallowxperm: {0:7}    Dontauditxperm:  {1:7}".format(
                p.auditallowxperm_count, p.dontauditxperm_count))
            print("  Initial SIDs:    {0:7}    Fs_use:          {1:7}".format(
                p.initialsids_count, p.fs_use_count))
            print("  Genfscon:        {0:7}    Portcon:         {1:7}".format(
                p.genfscon_count, p.portcon_count))
            print("  Netifcon:        {0:7}    Nodecon:         {1:7}".format(
                p.netifcon_count, p.nodecon_count))
        elif p.target_platform == "xen":
            print("  Initial SIDs:    {0:7}    Devicetreecon    {1:7}".format(
                p.initialsids_count, p.devicetreecon_count))
            print("  Iomemcon:        {0:7}    Ioportcon:       {1:7}".format(
                p.iomemcon_count, p.ioportcon_count))
            print("  Pcidevicecon:    {0:7}    Pirqcon:         {1:7}".format(
                p.pcidevicecon_count, p.pirqcon_count))

    for desc, component, expander in components:
        results = sorted(component.results())
        if not args.flat:
            print("\n{0}: {1}".format(desc, len(results)))
        for item in results:
            result = expander(item) if args.expand else item
            strfmt = "   {0}" if not args.flat else "{0}"
            print(strfmt.format(result))

except Exception as err:
    if args.debug:
        logging.exception(str(err))
    else:
        print(err)

    sys.exit(1)



# Authors: Karl MacMillan <kmacmillan@mentalrootkit.com>
#
# Copyright (C) 2006 Red Hat 
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import unittest
import sepolgen.refpolicy as refpolicy
import selinux

class TestIdSet(unittest.TestCase):
    def test_set_to_str(self):
        s = refpolicy.IdSet(["read", "write", "getattr"])
        s = s.to_space_str().split(' ')
        s.sort()
        expected = "{ read write getattr }".split(' ')
        expected.sort()
        self.assertEqual(s, expected)
        s = refpolicy.IdSet()
        s.add("read")
        self.assertEqual(s.to_space_str(), "read")

class TestSecurityContext(unittest.TestCase):
    def test_init(self):
        sc = refpolicy.SecurityContext()
        sc = refpolicy.SecurityContext("user_u:object_r:foo_t")
    
    def test_from_string(self):
        context = "user_u:object_r:foo_t"
        sc = refpolicy.SecurityContext()
        sc.from_string(context)
        self.assertEqual(sc.user, "user_u")
        self.assertEqual(sc.role, "object_r")
        self.assertEqual(sc.type, "foo_t")
        self.assertEqual(sc.level, None)
        if selinux.is_selinux_mls_enabled():
            self.assertEqual(str(sc), context + ":s0")
        else:
            self.assertEqual(str(sc), context)
        self.assertEqual(sc.to_string(default_level="s1"), context + ":s1")

        context = "user_u:object_r:foo_t:s0-s0:c0-c255"
        sc = refpolicy.SecurityContext()
        sc.from_string(context)
        self.assertEqual(sc.user, "user_u")
        self.assertEqual(sc.role, "object_r")
        self.assertEqual(sc.type, "foo_t")
        self.assertEqual(sc.level, "s0-s0:c0-c255")
        self.assertEqual(str(sc), context)
        self.assertEqual(sc.to_string(), context)

        sc = refpolicy.SecurityContext()
        self.assertRaises(ValueError, sc.from_string, "abc")

    def test_equal(self):
        sc1 = refpolicy.SecurityContext("user_u:object_r:foo_t")
        sc2 = refpolicy.SecurityContext("user_u:object_r:foo_t")
        sc3 = refpolicy.SecurityContext("user_u:object_r:foo_t:s0")
        sc4 = refpolicy.SecurityContext("user_u:object_r:bar_t")

        self.assertEqual(sc1, sc2)
        self.assertNotEqual(sc1, sc3)
        self.assertNotEqual(sc1, sc4)

class TestObjecClass(unittest.TestCase):
    def test_init(self):
        o = refpolicy.ObjectClass(name="file")
        self.assertEqual(o.name, "file")
        self.assertTrue(isinstance(o.perms, set))

class TestAVRule(unittest.TestCase):
    def test_init(self):
        a = refpolicy.AVRule()
        self.assertEqual(a.rule_type, a.ALLOW)
        self.assertTrue(isinstance(a.src_types, set))
        self.assertTrue(isinstance(a.tgt_types, set))
        self.assertTrue(isinstance(a.obj_classes, set))
        self.assertTrue(isinstance(a.perms, set))

    def test_to_string(self):
        a = refpolicy.AVRule()
        a.src_types.add("foo_t")
        a.tgt_types.add("bar_t")
        a.obj_classes.add("file")
        a.perms.add("read")
        self.assertEqual(a.to_string(), "allow foo_t bar_t:file read;")

        a.rule_type = a.DONTAUDIT
        a.src_types.add("user_t")
        a.tgt_types.add("user_home_t")
        a.obj_classes.add("lnk_file")
        a.perms.add("write")
        # This test might need to go because set ordering is not guaranteed
        a = a.to_string().split(' ')
        a.sort()
        b = "dontaudit { foo_t user_t } { user_home_t bar_t }:{ lnk_file file } { read write };".split(' ')
        b.sort()
        self.assertEqual(a, b)

class TestTypeRule(unittest.TestCase):
    def test_init(self):
        a = refpolicy.TypeRule()
        self.assertEqual(a.rule_type, a.TYPE_TRANSITION)
        self.assertTrue(isinstance(a.src_types, set))
        self.assertTrue(isinstance(a.tgt_types, set))
        self.assertTrue(isinstance(a.obj_classes, set))
        self.assertEqual(a.dest_type, "")

    def test_to_string(self):
        a = refpolicy.TypeRule()
        a.src_types.add("foo_t")
        a.tgt_types.add("bar_exec_t")
        a.obj_classes.add("process")
        a.dest_type = "bar_t"
        self.assertEqual(a.to_string(), "type_transition foo_t bar_exec_t:process bar_t;")


class TestParseNode(unittest.TestCase):
    def test_walktree(self):
        # Construct a small tree
        h = refpolicy.Headers()
        a = refpolicy.AVRule()
        a.src_types.add("foo_t")
        a.tgt_types.add("bar_t")
        a.obj_classes.add("file")
        a.perms.add("read")

        ifcall = refpolicy.InterfaceCall(ifname="allow_foobar")
        ifcall.args.append("foo_t")
        ifcall.args.append("{ file dir }")

        i = refpolicy.Interface(name="foo")
        i.children.append(a)
        i.children.append(ifcall)
        h.children.append(i)

        a = refpolicy.AVRule()
        a.rule_type = a.DONTAUDIT
        a.src_types.add("user_t")
        a.tgt_types.add("user_home_t")
        a.obj_classes.add("lnk_file")
        a.perms.add("write")
        i = refpolicy.Interface(name="bar")
        i.children.append(a)
        h.children.append(i)

class TestHeaders(unittest.TestCase):
    def test_iter(self):
        h = refpolicy.Headers()
        h.children.append(refpolicy.Interface(name="foo"))
        h.children.append(refpolicy.Interface(name="bar"))
        h.children.append(refpolicy.ClassMap("file", "read write"))
        i = 0
        for node in h:
            i += 1
        self.assertEqual(i, 3)
        
        i = 0
        for node in h.interfaces():
            i += 1
        self.assertEqual(i, 2)
        



# Authors: Karl MacMillan <kmacmillan@mentalrootkit.com>
#
# Copyright (C) 2006 Red Hat 
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 only
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import unittest
import sepolgen.refpolicy as refpolicy
import sepolgen.refparser as refparser
import sepolgen.policygen as policygen
import sepolgen.access as access

class TestAccessVector(unittest.TestCase):
    def test_init(self):
        # Default construction
        a = access.AccessVector()
        self.assertEqual(a.src_type, None)
        self.assertEqual(a.tgt_type, None)
        self.assertEqual(a.obj_class, None)
        self.assertTrue(isinstance(a.perms, refpolicy.IdSet))
        self.assertTrue(isinstance(a.audit_msgs, type([])))
        self.assertEqual(len(a.audit_msgs), 0)

        # Construction from a list
        a = access.AccessVector()
        a.src_type = "foo"
        a.tgt_type = "bar"
        a.obj_class = "file"
        a.perms.update(["read", "write"])

        l = access.AccessVector(['foo', 'bar', 'file', 'read', 'write'])
        self.assertEqual(a.src_type, l.src_type)
        self.assertEqual(a.tgt_type, l.tgt_type)
        self.assertEqual(a.obj_class, l.obj_class)
        self.assertEqual(a.perms, l.perms)

    def test_from_list(self):
        a = access.AccessVector()
        a.src_type = "foo"
        a.tgt_type = "bar"
        a.obj_class = "file"
        a.perms.update(["read", "write"])

        l = access.AccessVector()
        l.from_list(['foo', 'bar', 'file', 'read', 'write'])
        self.assertEqual(a.src_type, l.src_type)
        self.assertEqual(a.tgt_type, l.tgt_type)
        self.assertEqual(a.obj_class, l.obj_class)
        self.assertEqual(a.perms, l.perms)

    def test_to_list(self):
        a = access.AccessVector()
        a.src_type = "foo"
        a.tgt_type = "bar"
        a.obj_class = "file"
        a.perms.update(["read", "write"])

        l = a.to_list()
        self.assertEqual(l[0], "foo")
        self.assertEqual(l[1], "bar")
        self.assertEqual(l[2], "file")
        perms = l[3:]
        perms.sort()
        self.assertEqual(perms[0], "read")
        self.assertEqual(perms[1], "write")

    def test_to_string(self):
        a = access.AccessVector()
        a.src_type = "foo"
        a.tgt_type = "bar"
        a.obj_class = "file"
        a.perms.update(["read", "write"])

        first, second = str(a).split(':')
        self.assertEqual(first, "allow foo bar")
        second = second.split(' ')
        second.sort()
        expected = "file { read write };".split(' ')
        expected.sort()
        self.assertEqual(second, expected)

        first, second = a.to_string().split(':')
        self.assertEqual(first, "allow foo bar")
        second = second.split(' ')
        second.sort()
        expected = "file { read write };".split(' ')
        expected.sort()
        self.assertEqual(second, expected)

    def test_cmp(self):
        a = access.AccessVector()
        a.src_type = "foo"
        a.tgt_type = "bar"
        a.obj_class = "file"
        a.perms.update(["read", "write"])

        b = access.AccessVector()
        b.src_type = "foo"
        b.tgt_type = "bar"
        b.obj_class = "file"
        b.perms.update(["read", "write"])

        self.assertEqual(a, b)

        # Source Type
        b.src_type = "baz"
        self.assertNotEqual(a, b)
        self.assertTrue(a > b)

        b.src_type = "gaz"
        self.assertNotEqual(a, b)
        self.assertTrue(a < b)

        # Target Type
        b.src_type = "foo"
        b.tgt_type = "aar"
        self.assertNotEqual(a, b)
        self.assertTrue(a > b)

        b.tgt_type = "gaz"
        self.assertNotEqual(a, b)
        self.assertTrue(a < b)

        # Perms
        b.tgt_type = "bar"
        b.perms = refpolicy.IdSet(["read"])
        self.assertNotEqual(a, b)
        self.assertTrue(a > b)

        b.perms = refpolicy.IdSet(["read", "write", "append"])
        self.assertNotEqual(a, b)

        b.perms = refpolicy.IdSet(["read", "append"])
        self.assertNotEqual(a, b)
                         
class TestUtilFunctions(unittest.TestCase):
    def test_is_idparam(self):
        self.assertTrue(access.is_idparam("$1"))
        self.assertTrue(access.is_idparam("$2"))
        self.assertTrue(access.is_idparam("$123"))
        self.assertFalse(access.is_idparam("$123.23"))
        self.assertFalse(access.is_idparam("$A"))

    def test_avrule_to_access_vectors(self):
        rule = refpolicy.AVRule()
        rule.src_types.add("foo")
        rule.src_types.add("baz")
        rule.tgt_types.add("bar")
        rule.tgt_types.add("what")
        rule.obj_classes.add("file")
        rule.obj_classes.add("dir")
        rule.perms.add("read")
        rule.perms.add("write")

        avs = access.avrule_to_access_vectors(rule)
        self.assertEqual(len(avs), 8)
        comps = [("foo", "what", "dir"),
                 ("foo", "what", "file"),
                 ("foo", "bar", "dir"),
                 ("foo", "bar", "file"),
                 ("baz", "what", "dir"),
                 ("baz", "what", "file"),
                 ("baz", "bar", "dir"),
                 ("baz", "bar", "file")]
        status = [False] * 8
        for av in access.avrule_to_access_vectors(rule):
            self.assertEqual(av.perms, refpolicy.IdSet(["read", "write"]))
            for i in range(len(comps)):
                if comps[i][0] == av.src_type and \
                   comps[i][1] == av.tgt_type and \
                   comps[i][2] == av.obj_class:
                    status[i] = True

        for s in status:
            self.assertEqual(s, True)
                   

class TestAccessVectorSet(unittest.TestCase):
    def setUp(self):
        rule = refpolicy.AVRule()
        rule.src_types.add("foo")
        rule.src_types.add("baz")
        rule.tgt_types.add("bar")
        rule.tgt_types.add("what")
        rule.obj_classes.add("file")
        rule.obj_classes.add("dir")
        rule.perms.add("read")
        rule.perms.add("write")

        s = access.AccessVectorSet()
        avs = access.avrule_to_access_vectors(rule)
        for av in avs:
            s.add_av(av)
        self.s = s
    
    def test_init(self):
        a = access.AccessVectorSet()

    def test_iter(self):
        comps = [("foo", "what", "dir"),
                 ("foo", "what", "file"),
                 ("foo", "bar", "dir"),
                 ("foo", "bar", "file"),
                 ("baz", "what", "dir"),
                 ("baz", "what", "file"),
                 ("baz", "bar", "dir"),
                 ("baz", "bar", "file")]
        status = [False] * 8
        for av in self.s:
            self.assertEqual(av.perms, refpolicy.IdSet(["read", "write"]))
            for i in range(len(comps)):
                if comps[i][0] == av.src_type and \
                   comps[i][1] == av.tgt_type and \
                   comps[i][2] == av.obj_class:
                    status[i] = True

        for s in status:
            self.assertEqual(s, True)

    def test_len(self):
        self.assertEqual(len(self.s), 8)

    def test_list(self):
        a = access.AccessVectorSet()
        a.add("$1", "foo", "file", refpolicy.IdSet(["read", "write"]))
        a.add("$1", "bar", "file", refpolicy.IdSet(["read", "write"]))
        a.add("what", "bar", "file", refpolicy.IdSet(["read", "write"]))

        avl = a.to_list()
        avl.sort()

        test_l = [['what','bar','file','read','write'],
                  ['$1','foo','file','read','write'],
                  ['$1','bar','file','read','write']]
        test_l.sort()

        for a,b in zip(test_l, avl):
            self.assertEqual(len(a), len(b))
            for x,y in list(zip(a,b))[:3]:
                self.assertEqual(x, y)
            perms1 = a[3:]
            perms2 = b[3:]
            perms1.sort()
            perms2.sort()
            self.assertEqual(perms1, perms2)
                
        b = access.AccessVectorSet()
        b.from_list(avl)
        self.assertEqual(len(b), 3)
